/*
   Copyright (C) 2001,2002,2003,2004,2005,2006,2007 Keisuke Nishida
   Copyright (C) 2007,2008,2009,2010 Roger While

   This file is part of OpenCOBOL.

   The OpenCOBOL compiler is free software: you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   OpenCOBOL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with OpenCOBOL.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef CB_COBC_H
#define CB_COBC_H

#include <stdio.h>

#include "libcob.h"


#ifdef	ENABLE_NLS
#include "lib/gettext.h"
#define _(s)	gettext(s)
#define N_(s)	gettext_noop(s)
#else
#define _(s)	s
#define N_(s)	s
#endif

#define ABORT()			cobc_abort(__FILE__,__LINE__)

/* Source format defines */
#define CB_FORMAT_FIXED			0
#define CB_FORMAT_FREE			1

/* COPY extended syntax defines */
#define CB_REPLACE_LEADING		1
#define CB_REPLACE_TRAILING		2

/* Stringify macros */
#define COBSTRINGIFY(s)			#s
#define COBXSTRINGIFY(s)		COBSTRINGIFY(s)

/* ASSIGN clause interpretation */
#define CB_ASSIGN_MF			0	/* Micro Focus compatibility */
#define CB_ASSIGN_IBM			1U	/* IBM compatibility */
#define CB_ASSIGN_COBOL2002		2U	/* COBOL 2002 standard */

/* COMP/BINARY byte order */
#define CB_BYTEORDER_BIG_ENDIAN		0
#define CB_BYTEORDER_NATIVE		1U

/* Binary field sizes */
#define CB_BINARY_SIZE_1_2_4_8		0	/* 1,2,4,8 bytes */
#define CB_BINARY_SIZE_1__8		1U	/* 1,2,3,4,5,6,7,8 bytes */
#define CB_BINARY_SIZE_2_4_8		2U	/* 2,4,8 bytes */

/* Flex directive actions */
#define PLEX_ACT_IF			0
#define PLEX_ACT_ELSE			1U
#define PLEX_ACT_END			2U
#define PLEX_ACT_ELIF			3U

/* Flex value types */
#define PLEX_DEF_NONE			0
#define PLEX_DEF_LIT			1U
#define PLEX_DEF_NUM			2U
#define PLEX_DEF_DEL			3U

/* Operand operation type */
enum cb_operation_type {
	CB_OPERATION_READ,
	CB_OPERATION_WRITE,
	CB_OPERATION_ASSIGN
};

/* Config dialect support types */
enum cb_support {
	CB_OK,
	CB_WARNING,
	CB_ARCHAIC,
	CB_OBSOLETE,
	CB_SKIP,
	CB_IGNORE,
	CB_ERROR,
	CB_UNCONFORMABLE
};

struct cb_text_list {
	struct cb_text_list	*next;
	const char		*text;
};

struct cb_replace_list {
	struct cb_replace_list	*next;
	struct cb_text_list	*old_text;
	struct cb_text_list	*new_text;
	size_t			leading_trailing;
};

struct cb_define_struct {
	struct cb_define_struct	*next;
	char			*name;
	char			*value;
	unsigned int		deftype;
	int			sign;
	int			int_part;
	int			dec_part;
};

struct local_filename {
	struct local_filename	*next;
	char			*local_name;
	FILE			*local_fp;
};

struct filename {
	struct filename		*next;
	const char		*source;		/* foo.cob */
	const char		*preprocess;		/* foo.i */
	const char		*translate;		/* foo.c */
	const char		*trstorage;		/* foo.c.h */
	const char		*object;		/* foo.o */
	const char		*demangle_source;	/* foo */
	const char		*listing_file;		/* foo.lst */
	struct local_filename	*localfile;		/* foo.c.l[n].h */
	unsigned int		need_preprocess;
	unsigned int		need_translate;
	unsigned int		need_assemble;
};

struct cb_exception {
	const char	*name;			/* Exception name */
	const int	code;			/* Exception code */
	int		enable;			/* If turned on */
};

struct noreserve {
	struct	noreserve	*next;
	char			*noresword;
};


extern int	cb_source_format;

extern int	cb_display_sign;
extern int	cb_correct_numeric;

extern struct cb_exception	cb_exception_table[];

#define CB_EXCEPTION_NAME(id)	cb_exception_table[id].name
#define CB_EXCEPTION_CODE(id)	cb_exception_table[id].code
#define CB_EXCEPTION_ENABLE(id)	cb_exception_table[id].enable

#undef	CB_FLAG
#define	CB_FLAG(var,pdok,name,doc)	extern int var;
#include "flag.def"
#undef	CB_FLAG

#undef	CB_WARNDEF
#define	CB_WARNDEF(var,name,wall,doc)	extern int var;
#include "warning.def"
#undef	CB_WARNDEF

#undef	CB_OPTIM_DEF
#define	CB_OPTIM_DEF(x)			x,
enum cb_optim {
	COB_OPTIM_MIN,
#include "codeoptim.def"
	COB_OPTIM_MAX
};
#undef	CB_OPTIM_DEF

extern int			cb_id;
extern int			cb_attr_id;
extern int			cb_literal_id;
extern int			cb_field_id;
extern int			cobc_flag_main;
extern int			cb_flag_main;
extern int			cobc_wants_debug;
extern int			cb_stack_size;

extern int			errorcount;
extern int			warningcount;
extern int			warningopt;
extern int			no_physical_cancel;
extern int			alt_ebcdic;
extern int			optimize_defs[];

extern char			*cb_oc_build_stamp;
extern char			*cb_source_file;
extern int			cb_source_line;

extern const char		*cob_config_dir;

extern unsigned int		cobc_gen_listing;

extern const char		*demangle_name;
extern FILE			*cb_storage_file;
extern const char		*cb_storage_file_name;

extern char			**cb_saveargv;
extern int			cb_saveargc;

#if	0	/* Deprected */
extern FILE			*cb_depend_file;
extern struct cb_text_list	*cb_depend_list;
#endif

extern FILE			*cb_listing_file;
extern struct cb_text_list	*cb_include_list;
extern struct cb_text_list	*cb_extension_list;
extern struct cb_text_list	*cb_static_call_list;
extern struct cb_text_list	*cb_early_exit_list;
extern struct cb_define_struct	*cb_define_list;

extern struct cb_program	*current_program;
extern struct cb_statement	*current_statement;
extern struct cb_label		*current_section;
extern struct cb_label		*current_paragraph;
extern int			functions_are_all;

extern struct cb_text_list	*cb_text_list_add (struct cb_text_list *,
						   const char *);
extern void			*cobc_malloc (const size_t);
extern void			*cobc_strdup (const char *);
extern void			*cobc_realloc (void *, const size_t);

DECLNORET extern void		cobc_abort (const char *,
					    const int) COB_A_NORETURN;
DECLNORET extern void		cobc_too_many_errors (void);

extern size_t			cobc_check_valid_name (const char *,
						       const unsigned int);

/* config.c */

extern struct noreserve		*norestab;

#undef	CB_CONFIG_ANY
#undef	CB_CONFIG_INT
#undef	CB_CONFIG_STRING
#undef	CB_CONFIG_BOOLEAN
#undef	CB_CONFIG_SUPPORT

#define	CB_CONFIG_ANY(type,var,name)	extern type var;
#define	CB_CONFIG_INT(var,name)		extern int var;
#define	CB_CONFIG_STRING(var,name)	extern const char *var;
#define	CB_CONFIG_BOOLEAN(var,name)	extern int var;
#define	CB_CONFIG_SUPPORT(var,name)	extern enum cb_support var;

#include "config.def"

#undef	CB_CONFIG_ANY
#undef	CB_CONFIG_INT
#undef	CB_CONFIG_STRING
#undef	CB_CONFIG_BOOLEAN
#undef	CB_CONFIG_SUPPORT

extern unsigned int	cb_set_value (struct cb_define_struct *, const char *);
extern int		cb_load_std (const char *);
extern int		cb_load_conf (const char *, const int, const int);

/* preprocessor (in pplex.l, ppparse.y) */
extern FILE	*ppin;
extern FILE	*ppout;

extern int	pplex (void);
extern int	ppparse (void);
extern int	ppopen (const char *, struct cb_replace_list *);
extern int	ppcopy (const char *, const char *,
			struct cb_replace_list *);
extern void	pp_set_replace_list (struct cb_replace_list *);
extern void	ppparse_error (const char *);
extern void	plex_clear_vars (void);
extern void	plex_check_dangling (void);
extern void	plex_action_directive (const unsigned int, const unsigned int);

/* parser (in scanner.l, parser.y) */
extern FILE			*yyin;
extern FILE			*yyout;
extern char			*cobc_glob_line;
extern struct cb_define_struct	*cb_setvar_list;

extern int	yylex (void);
extern int	yyparse (void);
extern void	cobc_init_scanner (void);

/* typeck.c */
extern size_t	sending_id;
extern size_t	suppress_warn;
extern void	cobc_init_typeck (void);

/* reserved.c */
extern void	cobc_init_reserved (void);

/* error.c */
extern void		cb_warning (const char *, ...) COB_A_FORMAT12;
extern void		cb_error (const char *, ...) COB_A_FORMAT12;
extern void		cb_plex_warning (const size_t,
					 const char *, ...) COB_A_FORMAT23;
extern void		cb_plex_error (const size_t,
				       const char *, ...) COB_A_FORMAT23;

extern unsigned int	cb_verify (const enum cb_support, const char *);

#endif /* CB_COBC_H */
